@@ -5,13 +5,14 @@ import sys
 import marshal
 
 import tokyo.cabinet as tc
+import tokyo.tyrant as tt
 
-from fields import *
+from properties import *
 
 
-class class_cached_property(object):
+class cached_classmethod(object):
     '''
-    Кеширующий classmethod
+    cached_classmethod
     '''
     def __init__(self, f):
         self.f = f
@@ -27,31 +28,31 @@ class Error(Exception):
 
 
 class Model(object):
-    '''
-    Если у документа нет 
-
-    '''
-    __filename__ = None
+    __file__ = None
+    __host__ = None
     key = None
     
-    def __init__(self, key=None, _raw_dict=None, **kwargs):
+    def __init__(self, key=None, _raw_data=None, **kwargs):
         if key is not None:
             self.key = key if isinstance(key, unicode) else key.decode('utf-8')
-        if _raw_dict:
-            self._data = _raw_dict
+        self._cached_data = {}
+        if _raw_data:
+            self._raw_data = _raw_data
         else:
-            self._data = {}
+            self._raw_data = {}
             for prop_name, value in kwargs.iteritems():
-                if prop_name not in self._fields:
-                    assert 0, 'Unknown field "%s" in model "%s"' % (
-                        prop_name, self.__class__.__name__)
+                if prop_name not in self._properties:
+                    raise Error('Unknown field "%s" in model "%s"' % (
+                        prop_name, self.__class__.__name__))
                 setattr(self, prop_name, value)
     
     def save(self):
+        for name, key in self._presave_sync_properties:
+            if key in self._cached_data:
+                setattr(self, name, self._cached_data[key])
         if self.key is None:
             self.key = unicode(self.db.uid())
-        #~ print 'self.db[%s] = %s' % (self.key, self._data)
-        self.db[self.key] = self._data
+        self.db[self.key] = self._raw_data
         
     @classmethod
     def get(cls, key):
@@ -70,61 +71,69 @@ class Model(object):
 
     def delete(self):
         if not self.key:
-            assert 0, 'the object is not saved'
+            raise Error('the object is not saved')
         del self.db[self.key]
-        self._data = {}
+        self._raw_data = {}
+        self._cached_data = {}
 
-    @class_cached_property
+    @cached_classmethod
     def db(cls):
-        '''Отложенное подключение к бд'''
-        if not cls.__filename__:
-            assert 0, 'set __filename__ in model "%s"' % cls.__name__
-        db = tc.TDB()
-        db.open(cls.__filename__, tc.HDBOWRITER | tc.HDBOCREAT)
-        for prop in (cls.__dict__[n] for n in cls._fields):
-            #~ print prop.__class__.__name__
+        '''Lazy db connection'''
+        if cls.__file__:
+            db = tc.TDB()
+            db.open(cls.__file__, tc.HDBOWRITER | tc.HDBOCREAT)
+        elif cls.__host__:
+            host, port = cls.__host__.split(':')
+            db = tt.RTDB()
+            db.open(host=host, port=int(port))
+        else:
+            raise Error('set __file__ or __host__ in model "%s"' % 
+                cls.__name__)
+        for prop in (cls.__dict__[n] for n in cls._properties):
             if prop.index:
-                #~ print prop.__class__.__name__, prop.INDEX
                 if not prop.INDEX:
-                    raise Error('field "%s" is not indexed' % prop.__class__.__name__)
-                #~ print 'cls.db.setindex(%s, %s)' % (prop.key, prop.INDEX['set'])
+                    raise Error('field "%s" is not indexed' % 
+                        prop.__class__.__name__)
                 db.setindex(prop.key, prop.INDEX['set'])
             else:
                 try:
-                    db.setindex(prop.key, tc.TDBITVOID)
+                    db.setindex(prop.key, tt.TDBITVOID)
                 except (tc.Error, AttributeError):
                     pass
         return db
     
-    @class_cached_property
-    def _fields(cls):
+    @cached_classmethod
+    def _properties(cls):
         '''
         {prop_attr_name: save_key_name}
         '''
-        def rec_bases(cls):
-            bases = list(cls.__bases__) or []
-            ret = []
-            for base in bases:
-                for c in rec_bases(base):
-                    if c not in ret and c not in bases:
-                        ret.append(c)
-            ret.extend(bases)
-            return ret
         ret = {}
         for k, prop in cls.__dict__.items():
-            if isinstance(prop, KeyField):
+            if isinstance(prop, KeyProperty):
                 ret[k] = None
-            if BaseField in rec_bases(prop.__class__):
+            if issubclass(prop.__class__, BaseProperty):
                 if not prop.key:
                     prop.key = k
                 if prop.key in ret.values():
-                    assert 0, 'multiply key "%s" in model "%s"' % (prop.key, cls.__name__)
+                    raise Error('multiply key "%s" in model "%s"' % (prop.key, cls.__name__))
                 ret[k] = prop.key
         return ret
-        
+
+    @cached_classmethod
+    def _presave_sync_properties(cls):
+        '''
+        {prop_attr_name: save_key_name}
+        '''
+        ret = []
+        for name, key in cls._properties.iteritems():
+            prop = cls.__dict__[name]
+            if key and prop.PRESAVE_SYNC:
+                ret.append((name, key))
+        return ret
+
     def __unicode__(self):
         items = []
-        for k in ['key'] + self._fields.keys():
+        for k in ['key'] + self._properties.keys():
             v = getattr(self, k)
             if isinstance(v, str):
                 v = "'%s'" % v
@@ -158,12 +167,12 @@ class Filter(object):
         try:
             prop = self.model.__dict__[field]
         except KeyError:
-            assert 0, 'unknown field: "%s"' % field
+            raise Error('unknown field: "%s"' % field)
         try:
             condition = prop.get_condition(condition_abbr)
         except KeyError:
-            assert 0, 'unknown condition: "%s" for "%s" field' % (
-                condition_abbr, field)
+            raise Error('unknown condition: "%s" for "%s" field' % (
+                condition_abbr, field))
         queue = self.queue[:]
         queue.append((prop.key, condition, prop.dumps(value)))
         return Filter(self.model, queue, self._order, self._limit)
@@ -176,12 +185,11 @@ class Filter(object):
         try:
             prop = self.model.__dict__[field]
         except KeyError:
-            assert 0, 'unknown field: "%s"' % field
+            raise Error('unknown field: "%s"' % field)
         try:
-            #~ print prop.__class__.__name__, prop.key
             order = (prop.key, prop.get_order(direction))
         except KeyError:
-            assert 0, 'not ordered field: "%s"' % field
+            raise Error('not ordered field: "%s"' % field)
         return Filter(self.model, self.queue, order, self._limit)
         
     def __getitem__(self, key):
@@ -202,7 +210,6 @@ class Filter(object):
         for args in self.queue:
             q.filter(*args)
         if self._order:
-            #~ print 'order:', self._order
             q.sort(*self._order)
         if self._limit:
             q.limit(*self._limit)
@@ -210,7 +217,7 @@ class Filter(object):
 
     def __iter__(self):
         for k in self._query().search():
-            yield self.model(k, _raw_dict=self.model.db[k])
+            yield self.model(k, _raw_data=self.model.db[k])
     
     def count(self):
         return len(self._query().search())
@@ -228,48 +235,6 @@ class Filter(object):
             self.queue, self._order, self._limit)
 
 
-class A(Model):
-    number      = IntegerField(key='n', index=True)
-    fnum        = FloatField()
-    name        = UnicodeField(index=True)
-    birthday    = DateField(key='b', index=True)
-    created     = DateTimeField(index=True)
-    #~ tst         = ByteStringField(index=True)
-    
-    __filename__ = '/home/imbolc/0/var/otct.test.tct'
-
-
-#~ A.db.clear()
-
-#~ for i in xrange(100):
-    #~ A(
-        #~ number  = i,
-        #~ name    = u'ЫЫЫ',
-        #~ fnum    = -1.0 / (i + 1),
-        #~ birthday= date(1900 + i, 05, 06),
-        #~ created = datetime.now()
-    #~ ).save()
-
-#~ print '--- 2'
-#~ a.save()
-#~ print a
-
-#~ A.db.optimize()
-#~ print A.db.uid()
-
-#~ f = A.filter('number >=', 10)
-#~ f = f.filter('birthday <', date(1930, 1, 1))
-#~ print f.count()
-#~ for obj in f.order('fnum'):
-#~ for obj in f.order('-number'):
-    #~ print obj
-
-#~ f.delete()
-
-#~ a = A.get(6)
-#~ print a
-#~ a.delete()
-
 if __name__ == "__main__":
     import doctest
     doctest.testmod()
